from sf2utils.sf2parse import Sf2File
from sf2utils.sf2parse import Sf2Sample


import audioop
import logging
import sys
import wave
from contextlib import closing

from pathlib import Path
import subprocess

def raw_sample_data_loop(self):
    """return native endian linear buffer of samples (as expected by python wavfile)"""

    higher_part = self.sf2parser.read(self.loop_duration * 2, pos=self.smpl_offset + self.start * 2 + self.start_loop * 2)

    # soundfont smpl samples are packed as 16bits little endian, switch order if our system is big endian
    if sys.byteorder == 'big':
        higher_part = audioop.byteswap(higher_part, 2)

    # in 16bits only, return the top 16bits
    if self.sm24_offset is None:
        return higher_part

    # else read the complementary 8bits
    lower_part = self.sf2parser.read(self.loop_duration, pos=self.sm24_offset + self.start + self.start_loop)

    # and merge the result
    result = bytearray(self.loop_duration * 3)
    for idx in range(self.loop_duration):
        result[idx * 3: idx * + 1] = higher_part[idx * 2: idx * 2 + 1]
        result[idx * 3 + 2] = lower_part[idx]

    return result
Sf2Sample.raw_sample_data_loop = raw_sample_data_loop

def raw_sample_data_preloop(self):
    """return native endian linear buffer of samples (as expected by python wavfile)"""

    higher_part = self.sf2parser.read(self.start_loop * 2, pos=self.smpl_offset + self.start * 2)

    # soundfont smpl samples are packed as 16bits little endian, switch order if our system is big endian
    if sys.byteorder == 'big':
        higher_part = audioop.byteswap(higher_part, 2)

    # in 16bits only, return the top 16bits
    if self.sm24_offset is None:
        return higher_part

    # else read the complementary 8bits
    lower_part = self.sf2parser.read(self.start_loop, pos=self.sm24_offset + self.start)

    # and merge the result
    result = bytearray(self.start_loop * 3)
    for idx in range(self.start_loop):
        result[idx * 3: idx * + 1] = higher_part[idx * 2: idx * 2 + 1]
        result[idx * 3 + 2] = lower_part[idx]

    return result
Sf2Sample.raw_sample_data_preloop = raw_sample_data_preloop

def raw_sample_data_postloop(self):
    """return native endian linear buffer of samples (as expected by python wavfile)"""

    higher_part = self.sf2parser.read((self.duration - self.end_loop) * 2, pos=self.smpl_offset + self.start * 2 + self.end_loop * 2)

    # soundfont smpl samples are packed as 16bits little endian, switch order if our system is big endian
    if sys.byteorder == 'big':
        higher_part = audioop.byteswap(higher_part, 2)

    # in 16bits only, return the top 16bits
    if self.sm24_offset is None:
        return higher_part

    # else read the complementary 8bits
    lower_part = self.sf2parser.read((self.duration - self.end_loop), pos=self.sm24_offset + self.start + self.end_loop)

    # and merge the result
    result = bytearray((self.duration - self.end_loop) * 3)
    for idx in range((self.duration - self.end_loop)):
        result[idx * 3: idx * + 1] = higher_part[idx * 2: idx * 2 + 1]
        result[idx * 3 + 2] = lower_part[idx]

    return result
Sf2Sample.raw_sample_data_postloop = raw_sample_data_postloop

def raw_sample_data_range(self, start, end):
    """return native endian linear buffer of samples (as expected by python wavfile)"""
    
    # make sure the range asked for is sane
    if start < 0:
        start = 0
    if start > self.duration:
        start = self.duration
    if end > self.duration:
        end = self.duration
    if end < start:
        end = start
    duration = (end - start)

    higher_part = self.sf2parser.read(duration * 2, pos=self.smpl_offset + self.start * 2 + start * 2)

    # soundfont smpl samples are packed as 16bits little endian, switch order if our system is big endian
    if sys.byteorder == 'big':
        higher_part = audioop.byteswap(higher_part, 2)

    # in 16bits only, return the top 16bits
    if self.sm24_offset is None:
        return higher_part

    # else read the complementary 8bits
    lower_part = self.sf2parser.read(duration, pos=self.sm24_offset + self.start + start)

    # and merge the result
    result = bytearray(duration * 3)
    for idx in range(duration):
        result[idx * 3: idx * + 1] = higher_part[idx * 2: idx * 2 + 1]
        result[idx * 3 + 2] = lower_part[idx]

    return result
Sf2Sample.raw_sample_data_range = raw_sample_data_range

def export_loop(self, file):

    if self.smpl_offset is None:
        raise ValueError('no SMPL section found in Soundfont file, aborting sample export')

    # use of ''closing'' here allows the same code to be ran on python2 and python3
    with closing(wave.open(file, 'w')) as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(self.sample_width)
        wav_file.setframerate(self.sample_rate)
        wav_file.writeframes(self.raw_sample_data_loop())
Sf2Sample.export_loop = export_loop

def export_preloop(self, file):

    if self.smpl_offset is None:
        raise ValueError('no SMPL section found in Soundfont file, aborting sample export')

    # use of ''closing'' here allows the same code to be ran on python2 and python3
    with closing(wave.open(file, 'w')) as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(self.sample_width)
        wav_file.setframerate(self.sample_rate)
        wav_file.writeframes(self.raw_sample_data_preloop())
Sf2Sample.export_preloop = export_preloop

def export_postloop(self, file):

    if self.smpl_offset is None:
        raise ValueError('no SMPL section found in Soundfont file, aborting sample export')

    # use of ''closing'' here allows the same code to be ran on python2 and python3
    with closing(wave.open(file, 'w')) as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(self.sample_width)
        wav_file.setframerate(self.sample_rate)
        wav_file.writeframes(self.raw_sample_data_postloop())
Sf2Sample.export_postloop = export_postloop

def export_range(self, file, start, end):

    if self.smpl_offset is None:
        raise ValueError('no SMPL section found in Soundfont file, aborting sample export')

    # use of ''closing'' here allows the same code to be ran on python2 and python3
    with closing(wave.open(file, 'w')) as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(self.sample_width)
        wav_file.setframerate(self.sample_rate)
        wav_file.writeframes(self.raw_sample_data_range(start, end))
Sf2Sample.export_range = export_range


def split(mod_name, sample):
    tick_length = float(sample.sample_rate) / 60.0
    name_base = sample.name
    data_updates_prefix = f"""
instrument = table.insert(data.raw["programmable-speaker"]["programmable-speaker"].instruments, {"{"}
    name = "{mod_name}",
    notes =
    {"{"}
"""
    data_updates_suffix = """
    }
})
"""
    info_json = f"""{"{"}
  "name": "{mod_name}",
  "version": "1.0.0",
  "title": "{mod_name}",
  "author": "Illiander",
  "description": "Generated soundpack",
  "factorio_version": "1.1",
  "dependencies": []
{"}"}"""
    mod_folder = f"mods/{mod_name}/"
    soundfile_folder = f"{mod_folder}sound/"
    Path(soundfile_folder).mkdir(parents=True, exist_ok=True)
    
    with open(f"{mod_folder}info.json", 'w') as info:
        info.write(info_json)
    
    data_update_lines = []

    data_update_lines.append(factorio_mod_write_sample(sample, mod_name, "full-note", soundfile_folder, 0, sample.duration))
    
    segment_start = 0
    segment_end = tick_length
    segment_id = 0
    while segment_end <= sample.duration:
        data_update_lines.append(factorio_mod_write_sample(sample, mod_name, f"note-{segment_id}", soundfile_folder, int(segment_start), int(segment_end)))
        segment_id += 1
        segment_start = segment_end
        segment_end += tick_length
    
    data_updates = data_updates_prefix
    for i in data_update_lines:
        data_updates += i
    data_updates = data_updates[0:-2]
    data_updates += data_updates_suffix
    with open(f"{mod_folder}data-updates.lua", 'w') as info:
        info.write(data_updates)

#{ name = "alarm-1",  sound = { filename = "__Soundpack-alerts-and-sirens__/sound/alarm_1.ogg" }},
#ffmpeg -i my_sound_file.wav my_sound_file.ogg

def factorio_mod_write_sample(sample, mod_name, sound_name, soundfile_folder, start, end):
    data_line = f'        {"{"} name = "{sound_name}",  sound = {"{"} filename = "__{mod_name}__/sound/{sound_name}.ogg" {"}}"},\n'
    soundfile_name = f"{soundfile_folder}{sound_name}"
    sample.export_range(f"{soundfile_name}.wav", start, end)
    subprocess.run(["ffmpeg", "-i", f"{soundfile_name}.wav", f"{soundfile_name}.ogg"])
    Path(f"{soundfile_name}.wav").unlink(missing_ok=True)
    
    return data_line

#SGM-v2.01-YamahaGrand-Guit-Bass-v2.7.sf2
#FluidR3_GM/FluidR3_GM.sf2
#Creative (emu10k1)8MBGMSFX.SF2
#16.5mg_gm_gs_mt32_v2.51_bank.sf2
with open('soundfonts/16.5mg_gm_gs_mt32_v2.51_bank.sf2', 'rb') as sf2_file:
    sf2 = Sf2File(sf2_file)
    print(sf2.instruments[23].samples[5])
    print(sf2.instruments[23].samples[5].name)
    print(sf2.instruments[23].name)
    
    #for i,j in zip(sf2.samples, range(len(sf2.samples))):
    #    if "piano" in i.name.lower():
            #print(j)
    #        print(i.name)

    #for i in sf2.instruments[148].samples:
    #    print(i.name)
    
    monos = 0
    lefts = 0
    rights = 0
    errors = 0
    highest_sample_rate = 0
    max_duration = 0
    longest = None
    for i in sf2.instruments:
        for j in i.samples:
            sec_dur = j.duration/j.sample_rate
            if sec_dur > max_duration:
                max_duration = sec_dur
                longest = j
            highest_sample_rate = max(highest_sample_rate, j.sample_rate)
    #print(max_duration)
    #print(longest)
    #print(highest_sample_rate)
    #longest.export("dummy.wav")
        
    #sf2.instruments[23].samples[5].export("dummy.wav")
    #sf2.instruments[23].samples[5].export_loop("dummy_loop.wav")
    #sf2.instruments[23].samples[5].export_preloop("dummy_preloop.wav")
    #sf2.instruments[23].samples[5].export_postloop("dummy_postloop.wav")
    #sf2.instruments[23].samples[5].export_range("dummy_one.wav", 0, 10000)
    #sf2.instruments[23].samples[5].export_range("dummy_two.wav", 10000, 999999999999)


    mod_name = "testing-note"
    split(mod_name, sf2.instruments[23].samples[5])




















